:-module(xml2pro,[xml2pro/3]).

%-------------------------------------------------------------------------
%
%   XML -> PROLOG
%
%-------------------------------------------------------------------------

%---------------------------------------------------------------------------
%
%                        Mestrado em Informtica
%                       Jorge Manuel Neves Coelho
%
%        Data de incio do desenvolvimento: 25/09/2001
%        Data ltima alterao: 11/10/2001
%
%  Descrio:
%  ---------- 
%  Gerao de termo prolog a partir de documento XML e com base no seu DTD.
%
%---------------------------------------------------------------------------

%--------------------------------------------------------
% xml2pro(X,Y,Z) - predicado principal
%
% X - nome do ficheiro XML.
% Y - nome do ficheiro com o DTD.
% Z - termo gerado a partir das variveis anteriores.
%--------------------------------------------------------

xml2pro(FichXML,FichDTD,TermoP):-
	load_xml_file(FichXML,TermoX),
	new_dtd(nome_dtd,DTD), 
	load_dtd(DTD,FichDTD),
	processa(TermoX,DTD,TermoP).

%--------------------------------------------------------
% processa(X,Y,Z) - processamento e criao do termo.
%
% X - lista com o contedo do ficheiro XML.
% Y - lista com o contedo do DTD.
% Z - termo gerado aps processamento.
%--------------------------------------------------------

processa([element(X1,L2,L3)],DTD,Termo):-
	minusculas(X1,X2),
	dtd_property(DTD,element(X3,_,Tipo)),
	minusculas(X3,X4),
	X2==X4,
	processa_tipo(Tipo,DTD,X2,L3,Lista),
	dtd_property(DTD,attributes(Xattr,ListaAtr)),
	minusculas(Xattr,Xattrm),
	Xattrm==X2,
        valida_atributos(X2,ListaAtr,L2,NAtr),
	append([X2,NAtr],Lista,LFinal),
	Termo =..LFinal.

%--------------------------------------------------------
% processa_tipo(X,Y,Z,U,V) - processamento e validao do tipo do elemento
%
% X - tipo simples ou composto.
% Y - contedo do DTD.
% Z - nome do elemento cujo tipo est a ser analisado.
% U - lista com o contedo a ser validado.
% V - resultado da anlise do tipo (usado na construco do termo).
%--------------------------------------------------------

processa_tipo(empty,_,_,[],[]):-!.
processa_tipo('#pcdata',_,_,L,L):- !.
processa_tipo('#cdata',_,_,L,L):- !.
processa_tipo(Sub,DTD,_,L,LSub):- processa_composto(Sub,DTD,L,_,LSub). 

%--------------------------------------------------------
% verifica_pcdata(X) - verifica se os codigos ASCII na string representada
%                   por X estao entre determinados valores.
%
% X - lista com uma string.
%--------------------------------------------------------

verifica_pcdata([Str]):-
	atom_chars(Str,X),
	verifica_charpc(X).

verifica_charpc([]).
verifica_charpc([X|R]):-
	char_code(X,C),
	1 =< C,
	C =< 255,
	C \= 60,
	C \= 62,
	verifica_charpc(R).

%--------------------------------------------------------
% verifica_cdata(X) - verifica se os codigos ASCII na string representada
%                   por X estao entre determinados valores.
%
% X - lista com uma string.
%--------------------------------------------------------

verifica_cdata([Str]):-
	atom_chars(Str,X),
	verifica_charc(X).

verifica_charc([]).
verifica_charc([X|R]):-
	char_code(X,C),
	1 =< C,
	C =< 255,
	verifica_charc(R).

%--------------------------------------------------------
% processa_composto(X,Y,Z,U,V) - processamento de um tipo composto por sub-elementos.
%
% X - tuplo com os sub-elementos a tratar.
% Y - contedo do DTD.
% Z - lista de elementos (representao do ficheiro XML).
% U - resto da lista de elementos depois do processamento do tuplo de elementos.
% V - Lista com o termo resultante do processamento do tuplo.
%
% Notas:
% -----
%      ->  Porque as 'tags' no esto sempre na mesma forma (minusculas/maiusculas)
%          no XML e no DTD, todas as 'tags' so convertidas para minusculas
%          para proceder  unificao.
%      ->  Certo tipo de 'tags' no esto no DTD sobre a forma de 'atoms' (p.ex. ?xpto), 
%          por isso  necessria uma converso prvia com term_to_atom/2. 
%--------------------------------------------------------

% Processamento de um tuplo de dois ou mais elementos simples.
processa_composto((X,Y),DTD,LElem,RElem,LTerm):-
	processa_composto((X),DTD,LElem,RE,LTerm1),
	processa_composto((Y),DTD,RE,RElem,LTerm2),
	append(LTerm1,LTerm2,LTerm),!.

% Processamento de um tuplo com um s elemento simples.
processa_composto((X),DTD,[element(X1,Atr,T)|RElem],RElem,[Termo]):-
	minusculas(X1,X2),
	term_to_atom(X,Xt),
	minusculas(Xt,Xm),
	X2==Xm,
	dtd_property(DTD,element(X3,_,Tipo)),
	minusculas(X3,X4),
	X2==X4,
	processa_tipo(Tipo,DTD,X2,T,LT),!,
	dtd_property(DTD,attributes(Xattr,ListaAtr)),
	minusculas(Xattr,Xattrm),
	Xattrm==X2,
        valida_atributos(X2,ListaAtr,Atr,NAtr),
	append([X,NAtr],LT,LFinal),
        Termo=..LFinal.


% Processamento de um tuplo de repeties no fim do ficheiro XML.
processa_composto(*(_),_,[],[],[]):-!.


% Processamento de um tuplo de repeties nao encontrado.
processa_composto(*((X,_)),_,[element(X1,Atr,T)|RElem],[element(X1,Atr,T)|RElem],[]):-
	minusculas(X1,X2),
	term_to_atom(X,Xt),
	minusculas(Xt,Xm),
	X2\=Xm,!.


% Processamento de um tuplo de repeties e criao da respectiva lista de elementos.
processa_composto(*((X,Y)),DTD,[element(X1,Atr,T)|RElem],RElemF,[LTermo]):-
	processa_composto((X,Y),DTD,[element(X1,Atr,T)|RElem],RElemT,Termo1),
	processa_composto_rep((X,Y),DTD,RElemT,RElemF,Termo2),
	append(Termo1,Termo2,LTermo),!.

% Processamento de um tuplo de repeties sem mais elementos correspondentes.
processa_composto(*(X),_,[element(X1,Atr,T)|RElem],[element(X1,Atr,T)|RElem],[]):-
	minusculas(X1,X2),
	term_to_atom(X,Xt),
	minusculas(Xt,Xm),
	X2\=Xm,!. % aqui tem que chamar predicado que permita verificar se  um dos...

% Processamento de um tuplo de repeties e criao da respectiva lista de elementos.
processa_composto(*(X),DTD,[element(X1,Atr,T)|RElem],RElemF,[TermoF]):-
	minusculas(X1,X2),
	term_to_atom(X,Xt),
	minusculas(Xt,Xm),
	X2==Xm, 
	dtd_property(DTD,element(X3,_,Tipo)),
	minusculas(X3,X4),
	X2==X4,
	processa_tipo(Tipo,DTD,X2,T,LT),!,
	dtd_property(DTD,attributes(Xattr,ListaAtr)),
	minusculas(Xattr,Xattrm),
	Xattrm==X2,
        valida_atributos(X2,ListaAtr,Atr,NAtr),
	processa_composto_rep(X,DTD,RElem,RElemF,Termo2),
	Termo1=..[X,NAtr,LT],
	append([Termo1],Termo2,TermoF).


% Processamento de um tuplo opcional no fim do ficheiro XML.
processa_composto(?(_),_,[],[],[]).

% Processamento de um elemento opcional que no existe no ficheiro xml.
processa_composto(?(X,_),_,[element(X1,Atr,T)|RElem],[element(X1,Atr,T)|RElem],[]):-
	minusculas(X1,X2),
	term_to_atom(X,Xt),
	minusculas(Xt,Xm),
	X2\=Xm,!.


% Processamento de um elemento opcional que existe no ficheiro xml.
processa_composto(?((X,Y)),DTD,[element(X1,Atr,T)|RElem],RElem2,Termo):-
	minusculas(X1,X2),
	term_to_atom(X,Xt),
	minusculas(Xt,Xm),
	X2==Xm,
	processa_composto((X,Y),DTD,[element(X1,Atr,T)|RElem],RElem2,Termo),!.

% Processamento de um elemento opcional que no existe no ficheiro xml.
processa_composto(?(X),_,[element(X1,Atr,T)|RElem],[element(X1,Atr,T)|RElem],[]):-
	minusculas(X1,X2),
	term_to_atom(X,Xt),
	minusculas(Xt,Xm),
	X2\=Xm,!.

% Processamento de um elemento opcional que existe no ficheiro xml.
processa_composto(?(X),DTD,[element(X1,Atr,T)|RElem],RElem,[Termo]):-
	minusculas(X1,X2),
	term_to_atom(X,Xt),
	minusculas(Xt,Xm),
	X2==Xm,
	dtd_property(DTD,element(X3,_,Tipo)),
	minusculas(X3,X4),
	X2==X4,
	processa_tipo(Tipo,DTD,X2,T,LT),!,
	dtd_property(DTD,attributes(Xattr,ListaAtr)),
	minusculas(Xattr,Xattrm),
	Xattrm==X2,
        valida_atributos(X2,ListaAtr,Atr,NAtr),
	append([X,NAtr],LT,LFinal),
        Termo=..LFinal.

% Processamento de um tuplo de repeties e criao da respectiva lista de elementos.
processa_composto(+((X,Y)),DTD,[element(X1,Atr,T)|RElem],RElemF,[LTermo]):-
	processa_composto((X,Y),DTD,[element(X1,Atr,T)|RElem],RElemT,Termo1), % um elemento obrigatrio
	processa_composto_rep((X,Y),DTD,RElemT,RElemF,Termo2),
	append(Termo1,Termo2,LTermo).

% Processamento de um tuplo de uma ou mais repeties e criao da respectiva lista de elementos.
processa_composto(+(X),DTD,[element(X1,Atr,T)|RElem],RElemF,[TermoF]):-
	minusculas(X1,X2),
	term_to_atom(X,Xt),
	minusculas(Xt,Xm),
	X2==Xm, 
	dtd_property(DTD,element(X3,_,Tipo)),
	minusculas(X3,X4),
	X2==X4,
	processa_tipo(Tipo,DTD,X2,T,LT),!,
	dtd_property(DTD,attributes(Xattr,ListaAtr)),
	minusculas(Xattr,Xattrm),
	Xattrm==X2,
        valida_atributos(X2,ListaAtr,Atr,NAtr),
	processa_composto_rep(X,DTD,RElem,RElemF,Termo2),
	Termo1=..[X,NAtr,LT],
	append([Termo1],Termo2,TermoF).


% Processamento de um tuplo de ous.
processa_composto((X|Y),DTD,El,REl,LTer):-
	processa_composto_ou((X|Y),DTD,El,REl,LTer).

%--------------------------------------------------------
% processa_composto_ou(X,Y,Z,U,V) - processamento de um tipo composto por sub-elementos opcionais.
%
% X - tuplo com os sub-elementos a tratar.
% Y - contedo do DTD.
% Z - lista de elementos (representao do ficheiro XML).
% U - resto da lista de elementos depois do processamento do tuplo de elementos.
% V - Lista com o termo resultante do processamento do tuplo.
%--------------------------------------------------------

processa_composto_ou((X),DTD,[element(X1,Atr,T)|RElem],RElem,[Termo]):-
	minusculas(X1,X2),
	term_to_atom(X,Xt),
	minusculas(Xt,Xm),
	X2==Xm,
	dtd_property(DTD,element(X3,_,Tipo)),
	minusculas(X3,X4),
	X2==X4,
	processa_tipo(Tipo,DTD,X2,T,LT),!,
	dtd_property(DTD,attributes(Xattr,ListaAtr)),
	minusculas(Xattr,Xattrm),
	Xattrm==X2,
        valida_atributos(X2,ListaAtr,Atr,NAtr),
	append([X,NAtr],LT,LFinal),
        Termo=..LFinal.

processa_composto_ou((X|_),DTD,[element(X1,Atr,T)|RElem],RElem,[Termo]):-
	minusculas(X1,X2),
	term_to_atom(X,Xt),
	minusculas(Xt,Xm),
	X2==Xm,
	dtd_property(DTD,element(X3,_,Tipo)),
	minusculas(X3,X4),
	X2==X4,
	processa_tipo(Tipo,DTD,X2,T,LT),!,
	dtd_property(DTD,attributes(Xattr,ListaAtr)),
	minusculas(Xattr,Xattrm),
	Xattrm==X2,
        valida_atributos(X2,ListaAtr,Atr,NAtr),
	append([X,NAtr],LT,LFinal),
        Termo=..LFinal.

processa_composto_ou((X|Y),DTD,[element(X1,Atr,T)|RElem],RElemF,LTermo):-
	minusculas(X1,X2),
	term_to_atom(X,Xt),
	minusculas(Xt,Xm),
	X2\=Xm,
	processa_composto_ou((Y),DTD,[element(X1,Atr,T)|RElem],RElemF,LTermo).


%--------------------------------------------------------
% processa_composto_rep(X,Y,Z,U,V) - processamento de um tipo composto por sub-elementos que se repetem.
%
% X - tuplo com os sub-elementos a tratar.
% Y - contedo do DTD.
% Z - lista de elementos (representao do ficheiro XML).
% U - resto da lista de elementos depois do processamento do tuplo de elementos.
% V - Lista com o termo resultante do processamento do tuplo.
%--------------------------------------------------------

% Processamento de um elemento sem mais repeties e no fim do ficheiro XML.
processa_composto_rep(_,_,[],[],[]):-!.

% Processamento de um elemento de repeties, composto por sub-elementos.
processa_composto_rep((X,_),[element(X1,_,_)|_],_,_,[]):-
	minusculas(X1,X2),
	term_to_atom(X,Xt),
	minusculas(Xt,Xm),
	X2\=Xm.

% Processamento de um elemento de repeties, composto por sub-elementos.
processa_composto_rep((X,Y),DTD,[element(X1,Atr,T)|RElem],RElemF,LTermo):-
	minusculas(X1,X2),
	term_to_atom(X,Xt),
	minusculas(X,Xt),
	X2==Xt,
	processa_composto((X,Y),DTD,[element(X1,Atr,T)|RElem],RElem1,LTermo1),
	processa_composto_rep((X,Y),DTD,RElem1,RElemF,LTermo2),
	append(LTermo1,LTermo2,LTermo),!.

% Processamento de um elemento sem mais repetioes.
processa_composto_rep(X,_,[element(X1,Atr,T)|RElem],[element(X1,Atr,T)|RElem],[]):-
	minusculas(X1,X2),
	term_to_atom(X,Xt),
	minusculas(Xt,Xm),
	X2\=Xm,!. 

% Processamento de um elemento e criao da sua lista para integrar o termo final
processa_composto_rep(X,DTD,[element(X1,Atr,T)|RElem],RElemF,TermoF):-
	minusculas(X1,X2),
	term_to_atom(X,Xt),
	minusculas(Xt,Xm),
	X2==Xm, 
	dtd_property(DTD,element(X3,_,Tipo)),
	minusculas(X3,X4),
	X2==X4,
	processa_tipo(Tipo,DTD,X2,T,LT),!,
	dtd_property(DTD,attributes(Xattr,ListaAtr)),
	minusculas(Xattr,Xattrm),
	Xattrm==X2,
        valida_atributos(X2,ListaAtr,Atr,NAtr),
	processa_composto_rep(X,DTD,RElem,RElemF,Termo2),
	Termo1=..[X,NAtr,LT],
	append([Termo1],Termo2,TermoF).

%--------------------------------------------------------
% valida_atributos(X,Y,Z,V) - validao dos atributos de um elemento.
%
% X - nome do elemento.
% Y - lista de atributos declarada no DTD.
% Z - lista de atributos no ficheiro XML.
% V - lista de atributos para incluir no termo final
%
% Nota: pode ser eliminado se bem adaptado o valida_um.
%--------------------------------------------------------

valida_atributos(_,[],[],[]):-!.
valida_atributos(_,L1,L2,LAtrT):-
	valida_um(L2,L1,LAtrT).

%--------------------------------------------------------
% valida_um(X,Y,Z) - dada uma lista de atributos, retorna a lista de termos para ser integrada no termo final.
%
% X - lista de atributos a partir do ficheiro xml.
% Y - lista de atributos declarados no DTD.
% Z - lista de termos resultante.
%--------------------------------------------------------

valida_um([],_,[]).
valida_um([X=Y|L1],L2,[Termo|RTermos]):-
	minusculas(X,Xmin),
	member(Atr,L2),
	minusculas(Atr,MAtr),
        Xmin==MAtr,
	Termo =.. [attribute,X,Y],
	valida_um(L1,L2,RTermos).
    
%--------------------------------------------------------
% minusculas(X,Y) - a partir de uma string em maisculas, cria um "atom" constitudo pela string em minusculas.
%
% X - "string" de maisculas.
% Y - "atom" criado a partir da converso de X para uma "string" de minsculas.
%--------------------------------------------------------

minusculas(Mai,AMin2):-
	string_to_list(Mai,LMai),
	converte_min_mai(LMai,LMin),
	string_to_list(AMin,LMin),
	string_to_atom(AMin,AMin2).

converte_min_mai([],[]).
converte_min_mai([39|R1],R2):-!,
	converte_min_mai(R1,R2).
converte_min_mai([X|L1],[Y|L2]):-
	char_type(Y,to_lower(X)),
	converte_min_mai(L1,L2).

%--------------------------------------------------------
% TESTES
%-------------------------------------------------------- 
% new_dtd(veiculo,DTD), load_dtd(DTD,'auto.dtd'),dtd_property(DTD,X).
% new_dtd(veiculo,DTD), load_dtd(DTD,'auto2.dtd'),dtd_property(DTD,X).
% new_dtd(order,DTD), load_dtd(DTD,'book.dtd'),dtd_property(DTD,X).
% new_dtd(acord,DTD), load_dtd(DTD,'ACORD_PC_XML_Format_V1_0_0.dtd'),dtd_property(DTD,X).
% xml2pro('auto.xml','auto.dtd',X).
% xml2pro('auto-fake.xml','auto.dtd',X).
% xml2pro('auto2.xml','auto2.dtd',X).
% xml2pro('auto4.xml','auto4.dtd',X).
% xml2pro('book.xml','book.dtd',X). 
% xml2pro('endereco.xml','endereco.dtd',X). 
% xml2pro('ACORD_PC_XML_Format_V1_0_0.xml','ACORD_PC_XML_Format_V1_0_0.dtd',X). 
